import re
import typing
from abc import ABC
from functools import partial

ends_with = partial(lambda x, y: x.endswith(y))
starts_with = partial(lambda x, y: x.startswith(y))
sub_with = partial(lambda x, y, z: x.replace(y, z))
re_match = partial(lambda x, y: y.search(x) is not None)


class Action(ABC):

    def run(self, x):
        raise NotImplementedError


class Match(ABC):
    def __init__(self, matched_type=str):
        self.matched_type = matched_type

    def _type_check(self, x):
        return isinstance(x, self.matched_type)

    def run(self, x):
        raise NotImplementedError


class EndsWith(Match):

    def __init__(self, char_pattern: str, matched_type=str):
        super().__init__(matched_type)
        self.matched_type = matched_type
        self.char = char_pattern

    def run(self, x):
        if not self._type_check(x):
            return False
        return x.endswith(self.char)


class StartsWith(Match):

    def __init__(self, char_pattern: str, matched_type=str):
        super().__init__(matched_type)
        self.matched_type = matched_type
        self.char = char_pattern

    def run(self, x):
        if not self._type_check(x):
            return False
        return x.startswith(self.char)


class MatchesRegex(Match):

    def __init__(self, pattern: typing.Union[str, typing.Pattern], matched_type=str):
        super().__init__(matched_type)
        self.matched_type = matched_type
        self.char = pattern
        if isinstance(self.char, str):
            self.f = partial(re.search, pattern=self.char)
        else:
            self.f = self.char.search

    def run(self, x: str):
        if not self._type_check(x):
            return False
        return self.f(string=x) is not None


class Sub(Action):

    def __init__(self, old: typing.Union[str, typing.Pattern], new: str):
        self.old = old
        self.new = new
        if isinstance(self.old, str):
            self.f = partial(lambda x: x.replace(self.old, self.new))
        else:
            self.f = partial(lambda ptrn, repl, x: ptrn.sub(repl, x), ptrn=self.old, repl=self.new)

    def run(self, x: str):
        x = self.f(x=x)
        return x


class Drop(Action):

    def run(self, x):
        return None


class Return(Action):

    def __init__(self, return_value):
        self.return_value = return_value

    def run(self, x):
        return self.return_value


class Pattern(object):
    def __init__(self, match: Match, action: Action):
        self.match = match
        self.action = action

    def __call__(self, *args, **kwargs):
        return self.run(*args, **kwargs)

    def run(self, x: str, *args, **kwargs):
        if self.match.run(x):
            x = self.action.run(x)
            return x, True
        return x, False


"""
Composed Patterns
"""


def __regex_factory(template, *flags, **kwargs):
    return re.compile(template.format(**kwargs), *flags)


def __regex_factory_i_escape(template, pattern):
    return __regex_factory(template, re.IGNORECASE, pattern=re.escape(pattern))


__startswith_template = r"^({pattern})"
__endswith_template = r"({pattern})$"
__contains_token_template = r"\b({pattern})\b"

_contains_ibm = __regex_factory_i_escape(__contains_token_template, "ibm")
contains_ibm = Pattern(
        match=MatchesRegex(pattern=_contains_ibm),
        action=Return(return_value="ibm")
        )

_contains_oracle = __regex_factory_i_escape(__contains_token_template, "oracle")
contains_oracle = Pattern(
        match=MatchesRegex(pattern=_contains_oracle),
        action=Return(return_value="oracle")
        )

_startswith_accenture = __regex_factory_i_escape(__startswith_template, "accenture")
starts_with_accenture = Pattern(
        match=MatchesRegex(pattern=_startswith_accenture),
        action=Return(return_value="accenture")
        )

_contains_hewlett_packard = __regex_factory_i_escape(__contains_token_template, "hewlett packard|(hpe)")
contains_hewlett_packard = Pattern(
        match=MatchesRegex(pattern=_contains_hewlett_packard),
        action=Return(return_value="hewlett packard")
        )

PATTERNS = [
    contains_oracle, contains_ibm, starts_with_accenture, contains_hewlett_packard
    ]



from company_graph.func import PATTERNS
from typing import AnyStr, Tuple

class RuleMatcher:

    RULES = PATTERNS

    @classmethod
    def run(cls, x: str) -> Tuple[AnyStr, bool]:
        for rule in RuleMatcher.RULES:
            result, matched = rule.run(x)
            if matched:
                return result, True
        return x, False

    @classmethod
    def __contains__(cls, item: str) -> bool:
        for rule in RuleMatcher.RULES:
            result, matched = rule.run(x)
            if matched:
                return True
        return False



import pickle

import pkg_resources

from company_graph.normalizer import Preprocessor
from company_graph.rules import RuleMatcher


class CompanyGraph(object):

    def __init__(self, preprocessor=Preprocessor(), rules=RuleMatcher):
        self.graph = self.load_graph()
        self.preprocessor = preprocessor
        self.rules = rules

    def load_graph(self):

        data_path = pkg_resources.resource_filename('company_graph', 'data/graph.pkl')
        with open(data_path, "rb") as pfile:
            return pickle.load(pfile)

    def __contains__(self, item):
        if self.preprocessor:
            if isinstance(item, str):
                item = self.preprocessor(item)
        if self.graph.has_node(item):
            return True
        else:
            return item in self.rules

    def __call__(self, item):
        if self.preprocessor:
            if isinstance(item, str):
                item = self.preprocessor(item)

        item, was_matched = self.rules.run(item)
        if was_matched:
            if not self.graph.has_node(item):
                return item
            else:
                if isinstance(item, str):
                    return self.string2id(item)

        elif not self.graph.has_node(item):
            return None

        if isinstance(item, str):
            return self.string2id(item)

        else:
            return self.id2string(item)

        # Is it a company_id? If so follow redirects

    def string2id(self, item):

        # String lookups will have one id mapping made available

        route = list(self.graph[item])[0]
        return route

    def id2id(self, item):

        # Does the id have any redirects?

        # If not, return id
        has_node = self.graph.has_node(item)
        if not has_node:
            return item
        routes = self.graph[item]
        if not routes:
            return item

        # If so...
        # Is the gravity of this node greater than all redirects?

        current_gravity = self.graph.nodes[item].get('g', 0)
        competing_node = max(routes.items(), key=lambda x: x[1].get('g', 0))

        # Yes? Stop here
        if current_gravity >= competing_node[1].get('g', 0):
            return item

        # No? Follow greater gravity
        return self.id2id(competing_node[0])

    def id2string(self, item):

        # Does the id have any redirects?
        item = self.id2id(item)

        # Generate potential names via predecessors
        # Filter out redirects, these have a gravity attribute
        potential_names = list(filter(lambda x: x[1].get('n') is not None, self.graph.pred.get(item).items()))

        if not potential_names:
            return None

        else:
            return max(potential_names, key=lambda x: x[1]['n'])[0]



import re
import string
import typing
from functools import partial


def strip_match(s, pattern):
    return pattern.sub(" ", s)


def to_lowercase(s):
    return s.lower()


RE_PUNCTUATION = re.compile(r'([%s])+' % re.escape(string.punctuation))
RE_MULTIPLE_WHITESPACE = re.compile(r"( {2,})")
STOPWORDS = ['for', 'global', 'by', 'international', 'technologies', 'the', 'group', 'systems', 'of', 'corporation',
             'and', 'limited', 'llc', 'ltd', 'inc', 'company', 'consulting', 'at']
RE_STOPWORDS = re.compile("({})(?![A-z] ?)".format("|".join(STOPWORDS)), re.IGNORECASE)

STRIP_PUNCTUATION = partial(strip_match, pattern=RE_PUNCTUATION)
STRIP_MULTIPLE_WHITESPACE = partial(strip_match, pattern=RE_MULTIPLE_WHITESPACE)
STRIP_STOPWORDS = partial(strip_match, pattern=RE_STOPWORDS)
TO_LOWERCASE = to_lowercase
STRIP_STRING = lambda x: x.strip()


class Preprocessor(object):
    DEFAULT_STRING_PROCESSORS = [STRIP_STOPWORDS, STRIP_PUNCTUATION, STRIP_MULTIPLE_WHITESPACE, TO_LOWERCASE, STRIP_STRING]

    def __init__(self, *args: typing.Callable):
        self.string_processors = self.DEFAULT_STRING_PROCESSORS if not args else args

    def __call__(self, s):
        ps = s  # Keep original reference
        for sp in self.string_processors:
            if not ps:  # In case we've removed everything
                return s
            ps = sp(ps)
        ps = ps.strip()
        if not ps:
            return s
        return ps



from hypothesis import given, example
from hypothesis.strategies import text, characters
import csv
import pytest

from company_graph.normalizer import *

@pytest.fixture(scope='module')
def stopwords():
    with open('../company_names_stopwords.csv', 'r', newline='', encoding='utf-8') as f:
        reader = csv.reader(f)
        headers = next(reader)
        test_pairs_stopwords = [tuple(line) for line in reader]
    return test_pairs_stopwords


@given(text() | characters())
def test_strip_punc(s):
    stripped_s = STRIP_PUNCTUATION(s)
    assert len([p for p in string.punctuation if p in stripped_s]) == 0

@given(text() | characters())
@example("ABC  ABC ABC")
@example("A B  A  B  ")
def test_multi_ws(s):

    # Are these still present?
    copy_s = s

    assert STRIP_MULTIPLE_WHITESPACE(copy_s).find("  ") == -1

    # Did we remove desired whitespace?
    original_tokens = [token for token in s.split() if token != " "]
    stripped_tokens = STRIP_MULTIPLE_WHITESPACE(copy_s).split()
    assert original_tokens == stripped_tokens

def test_stopwords(stopwords):
    for raw, expected in stopwords:
        assert STRIP_STOPWORDS(raw) == expected

    s = ['ABC-ABC', 'ABC.ABC', '.ABC.ABC', 'ABC at ABC', 'ABC ABC', 'ABC  ABC  ']
    p = Preprocessor()
    for example in s:
        assert "abc abc" == p(example)





import pytest

@pytest.fixture(scope='module')
def cg():
    from company_graph.cg import CompanyGraph
    cg = CompanyGraph()
    return cg

def test_company2id(cg):
    assert cg('Cisco') == 1063
    assert cg('Cisco Talos') == 1063
    assert cg('Google') == 1441
    assert cg(2844978) == 'cisco'
    assert cg.id2id(2844978) == 1063

